/*
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef AVRESAMPLE_AARCH64_ASM_OFFSETS_H
#define AVRESAMPLE_AARCH64_ASM_OFFSETS_H

/* struct ResampleContext */
#define FILTER_BANK                     0x10
#define FILTER_LENGTH                   0x18
#define PHASE_SHIFT                     0x34
#define PHASE_MASK                      (PHASE_SHIFT + 0x04) // loaded as pair

#endif /* AVRESAMPLE_AARCH64_ASM_OFFSETS_H */
